from dataclasses import dataclass
from typing import Any, Dict, List
import logging
import json

from tokko_auth.util import beautiful_errors, auth_is_enable
from tokko_auth.exceptions import NamespaceNotMatch
from tokko_auth.helpers import AuthorizationHelper
from tokko_auth.models import User

# App settings
from tokko_auth import settings

# Local logger instance
logger = logging.getLogger(__name__)


@dataclass
class MinimalAuth:
    """Support Auth class"""

    token: str = None
    claims: Dict = None
    scopes: List = None
    namespaces: List = None

    def __str__(self) -> str:
        return json.dumps(
            {
                "token": self.token,
                "claims": self.claims,
                "scopes": self.scopes,
                "namespaces": self.namespaces,
            },
            indent=4,
        )


class AuthMiddlewareBase:
    """Auth Base Middleware"""
    i_am_testing: bool = None
    expected_exception = None
    auth_request_property = "tokko_auth"

    def __init__(self, get_response, i_am_testing=False):
        self.i_am_testing = i_am_testing
        self.get_response = get_response

    def before_view(self, request) -> Any:
        return request

    def has_auth_layer(self, request) -> bool:
        return (
            hasattr(request, self.auth_request_property)
            or not request.authorization.token
        )

    def after_view(self, request, response) -> Any:
        logger.warning(f'{self}. AfterView Callback not configured')
        return response

    def __call__(self, request):
        if auth_is_enable() or self.i_am_testing:
            try:
                request = self.before_view(request)
            except Exception as exception:
                return beautiful_errors(exception)
        response = self.get_response(request)
        if auth_is_enable() or self.i_am_testing:
            try:
                response = self.after_view(request, response)
            except Exception as exception:
                return beautiful_errors(exception)
        return response

    def __str__(self) -> str:
        return f"{self.__class__.__name__}: {self.__doc__}"


class UserRecoverMiddleware(AuthMiddlewareBase, AuthorizationHelper):
    """Recover JWT user"""

    populate_userinfo = settings.AUTH_USERINFO_ENABLE

    def before_view(self, request) -> Any:
        token = self.get_token(request.headers)
        user_data = {
            "token": token,
            "namespaces": self.namespaces(token),
            "claims": self.claims(token),
            "scopes": self.scopes(token),
        }
        if self.populate_userinfo:
            user_data.update(
                {"userinfo": self.user_properties(request.authorization.token)}
            )
        setattr(request, "user", User(**user_data))
        return request


class HasJWTMiddleware(AuthMiddlewareBase, AuthorizationHelper):
    """Request has JWT middleware"""

    def before_view(self, request) -> Any:
        token = self.get_token(request.headers)
        auth = MinimalAuth(
            token,
            claims=self.claims(token),
            scopes=self.scopes(token),
            namespaces=self.namespaces(token),
        )
        setattr(request, self.auth_request_property, auth)
        return request


class NamespaceAuthorizationMiddleware(AuthMiddlewareBase, AuthorizationHelper):
    """Authorization by Namespace middleware"""

    namespace = settings.AUTH_API_NAMESPACE

    def before_view(self, request) -> Any:
        token = self.get_token(request.headers)
        if not self.has_namespace(self.namespace, token) and auth_is_enable():
            raise NamespaceNotMatch
        return request



from typing import Callable, Union, Any, Dict
from dataclasses import dataclass, field
from os import environ as env

from requests.exceptions import (ConnectionError, TooManyRedirects, HTTPError, ConnectTimeout)
from jose.exceptions import JWTError
from jose import jwt
import requests

from tokko_auth.exceptions import (
    TokenNotFound,
    MustBeBearerToken,
    MustStartWithBearer,
    ProviderUnreachable,
    UnableDecodeToken,
    TokenIsExpired,
    IncorrectClaims,
    ProviderHttpError
)
from tokko_auth.constants import AUTH0, COGNITO

AUTH_STRATEGY = COGNITO


def _cognito_jwk_strategy() -> str:
    region = env.get("AWS_COGNITO_REGION")
    pool_id = env.get("AWS_COGNITO_USERS_POOL_ID")

    return f"https://cognito-idp.{region}.amazonaws.com/{pool_id}/.well-known/jwks.json"


def _auth0_jwk_strategy() -> str:
    auth_domain = env.get("AUTH0_DOMAIN")

    return f"https://{auth_domain}/.well-known/jwks.json"


STRATEGIES: Dict[str, Callable] = {
    AUTH0: _auth0_jwk_strategy,
    COGNITO: _cognito_jwk_strategy,
}


def build_jwk_url(strategy: str = AUTH_STRATEGY) -> str:
    try:

        if not isinstance(strategy, str):
            raise TypeError("Unsupported strategy type")

        _jwk_url = STRATEGIES[strategy]

        if callable(_jwk_url):
            _jwk_url = _jwk_url()

        return _jwk_url

    except KeyError:
        raise KeyError("Unsupported Build JWK strategy")


def get_jwk(jwk_url) -> dict:
    try:
        r = requests.get(jwk_url)

        if not r.status_code == 200:
            r.raise_for_status()

        content_type = r.headers.get("Content-Type", "")
        if "application/json" not in content_type:
            raise ProviderHttpError(description=f"Invalid content type. {content_type} is not supported",
                                    status_code=500)

        _jwk = r.json()

        return _jwk

    except (ConnectionError, TooManyRedirects, ConnectTimeout) as e:
        # It is connected?
        raise IOError(f"Transport error. {e}")
    except HTTPError as e:
        raise IOError(f"Protocol Error. {e}")


def default_issuer_strategy(token: str, **options):
    _jwk_url = options.get("jwk_url")
    _algorithms = options.get("algorithms")
    _audience = options.get("audience")

    if not all([_jwk_url, _audience, _algorithms]):
        raise KeyError("ISSUER, ALGORITHMS & AUDIENCE are required arguments")

    try:
        jwks = get_jwk(_jwk_url)

        unverified_header = jwt.get_unverified_header(token)
        rsa_key = {}

        for key in jwks["keys"]:

            if key["kid"] == unverified_header["kid"]:
                rsa_key = {
                    "kty": key["kty"],
                    "kid": key["kid"],
                    "use": key["use"],
                    "n": key["n"],
                    "e": key["e"],
                }
        if rsa_key:
            try:
                payload = jwt.decode(
                    token,
                    rsa_key,
                    algorithms=_algorithms,
                    audience=_audience,
                )

                return payload

            except jwt.ExpiredSignatureError:
                raise TokenIsExpired

            except jwt.JWTClaimsError as e:
                raise IncorrectClaims

            except Exception:
                raise UnableDecodeToken

        raise KeyError(f"No RSA key found or not match.")

    except (ConnectionError, ConnectTimeout, HTTPError):
        raise ProviderUnreachable

    except JWTError:
        raise UnableDecodeToken

    except Exception as e:
        raise KeyError(f"Token validation fails. {e}")


def default_permission_strategy(token: str):
    ...


def get_auth_token_from_headers(key_name: str = "Authorization", **headers) -> Union[str, None]:
    """
    Returns Auth token from headers
    ===
    """
    auth = headers.get(key_name)

    # Syntactic Token validations
    if not auth:
        raise TokenNotFound

    parts = auth.split()

    if len(parts) == 1 or len(parts) > 2:
        raise MustBeBearerToken

    # Semantic Token validations
    if not parts[0].lower() == "bearer":
        raise MustStartWithBearer

    return parts[1]


@dataclass(init=True)
class Issuer:

    jwk_url: Union[Callable, str]

    strategy_kwargs: Dict[str, Any] = field(default_factory=lambda: {})
    jwk_url_kwargs: Dict[str, Any] = field(default_factory=lambda: {})
    auto_init: bool = True
    token: str = ""

    JWT_ISSUER = 'issuer'
    JWT_PERMISSION = 'permissions'

    AVAILABLE_STRATEGIES = {
        JWT_ISSUER: default_issuer_strategy,
        JWT_PERMISSION: lambda: 1
    }

    def get_jwk_url(self) -> str:

        _jwk_url = self.jwk_url

        if not _jwk_url:
            raise ValueError

        if callable(_jwk_url):
            return _jwk_url(self.token, **self.jwk_url_kwargs)

        elif isinstance(_jwk_url, str):
            return _jwk_url

        raise TypeError

    def approve_or_deny(self, token=None, strategy=JWT_ISSUER):

        if strategy not in self.AVAILABLE_STRATEGIES.keys():
            raise KeyError

        _strategy_callable = self.AVAILABLE_STRATEGIES[strategy]

        token = token or self.token

        if not token:
            raise ValueError

        self.strategy_kwargs.update({
            "jwk_url": self.jwk_url
        })

        _strategy_callable(token, **self.strategy_kwargs)

        return True

    def __post_init__(self):
        if self.auto_init:
            self.approve_or_deny()



from os import environ


# Error codes
UNAUTHORIZED = 401
INTERNAL_ERROR = 500
GATEWAY_ERROR = 503

# AVAILABLE STRATEGIES
AUTH0 = 'auth0'
COGNITO = 'aws-cognito'

AUTH_STRATEGY = COGNITO

ALGORITHMS = environ.get("TOKKO_AUTH_ALGORITHMS",
                         # Pre deprecation support
                         environ.get("ALGORITHMS"))

API_AUDIENCE = environ.get("TOKKO_AUTH_API_AUDIENCE",
                           # Pre deprecation support
                           environ.get("API_AUDIENCE"))

AUTH_DOMAIN = environ.get("TOKKO_AUTH_PROVIDER_DOMAIN",
                          # Pre deprecation support
                          environ.get("AUTH0_DOMAIN"))



from django.http import HttpResponse

from django.conf import settings
from json import dumps as _

from tokko_auth.settings import (
    AUTH_FULL_DISABLED as DISABLED_AUTH,
    AUTH_BEAUTIFY_ERROR_RESPONSE as SHOW_ME_BEAUTIFUL_ERRORS,
    AUTH_ALLOW_PRODUCTIVE_SHUTDOWN as ALLOW_UNPROTECTED_PRODUCTION,
    AUTH_FAIL_SAFE_ENABLE as FAIL_SAFE
)

from tokko_auth.exceptions import AuthPermissionErrors


def is_auth_exception(exception):
    return isinstance(exception, AuthPermissionErrors)


def auth_is_enable() -> bool:
    """
    Stop rules
    ---

    Is enable when:
        + settings DEBUG is True & AUTH_FULL_DISABLED is True
        + settings DEBUG is False, AUTH_FULL_DISABLED is True and AUTH_ALLOW_PRODUCTIVE_SHUTDOWN is True
    """
    debug_is_true = getattr(settings, "DEBUG", False)
    return any(
        [
            debug_is_true and not DISABLED_AUTH,
            not DISABLED_AUTH and ALLOW_UNPROTECTED_PRODUCTION and not debug_is_true,
        ]
    )


def beautiful_errors(exception, forced=False):
    if not FAIL_SAFE:
        if SHOW_ME_BEAUTIFUL_ERRORS and is_auth_exception(exception) and not forced:
            return HttpResponse(
                _(
                    {
                        "code": exception.error.code,
                        "description": exception.error.description,
                        "status_code": exception.status_code,
                    }
                ),
                content_type="application/json",
                status=exception.status_code,
            )
        raise exception



from dataclasses import dataclass
from typing import Any
import json

from django.test import TestCase

from tokko_auth.middleware import (HasJWTMiddleware, AuthMiddlewareBase)
from tokko_auth.tests.crypto import (
    INVALID_TOKEN_NO_CLAIMS,
    VALID_TOKEN,
    INVALID_TOKEN_NO_SCOPES,
    PAYLOAD,
)
from tokko_auth.helpers import (
    AuthorizationHelper,
    MustBeBearerToken,
    InsufficientScopes,
    ScopesNotFound,
    ClaimsNotFound,
    MustStartWithBearer,
)
from tokko_auth.exceptions import TokenNotFound, UnableDecodeToken


@dataclass
class FakeRequest:
    """Fake Request just for test proposes"""
    headers: dict = None
    text: str = None

    @property
    def content_type(self) -> str:
        try:
            self.json()
            return 'application/json'
        except json.decoder.JSONDecodeError:
            return 'text/plain'

    def json(self) -> dict:
        return json.loads(self.text)


def get_response(request: FakeRequest) -> Any:
    return request.__doc__


class MiddlewareUnitTest(TestCase):
    """Middleware Unitary tests"""
    def setUp(self) -> None:
        self.valid_token = VALID_TOKEN

    def test_01_invoke_base_middleware(self):
        """Successful MiddlewareBase Call"""
        base_middleware = AuthMiddlewareBase(get_response, i_am_testing=True)
        _request = FakeRequest(headers={'Authorization': f'Bearer {self.valid_token}'})
        self.assertEqual(base_middleware(_request), FakeRequest.__doc__)

    def test_02_invoke_has_jwt_middleware(self):
        """Call MiddlewareBase without Token"""
        has_token_middleware = HasJWTMiddleware(get_response, i_am_testing=True)
        _request = FakeRequest(headers={'Authorization': 'Bearer ...'})
        response = has_token_middleware(_request)
        self.assertEqual(response.status_code, 401)
        _json_response = json.loads(response.content.decode())
        self.assertEqual(_json_response, UnableDecodeToken().as_json)

    def test_03_invoke_has_jwt_middleware(self):
        """Success MiddlewareBase call"""
        has_token_middleware = HasJWTMiddleware(get_response, i_am_testing=True)
        _request = FakeRequest(headers={'Authorization': f'Bearer {self.valid_token}'})
        self.assertEqual(has_token_middleware(_request), FakeRequest.__doc__)


class HelpersUnitTests(TestCase):
    """Helpers module unitary tests"""

    def setUp(self) -> None:
        # Authorization Helper
        self.auth = AuthorizationHelper()
        # Token
        self.valid_token = VALID_TOKEN
        self.invalid_token_no_scopes = INVALID_TOKEN_NO_SCOPES
        self.invalid_token_no_claims = INVALID_TOKEN_NO_CLAIMS

    def test_01_get_token_success(self):
        """Successful exec"""
        _t = self.auth.get_token({"Authorization": f"Bearer {self.valid_token}"})
        self.assertEqual(_t, self.valid_token)

    def test_02_get_token_error_token_not_found(self):
        """AuthHelper.get_token Raises TokenNotFound when token not found in headers"""
        with self.assertRaises(TokenNotFound):
            self.auth.get_token({})

    def test_03_get_token_error_should_start_with_bearer(self):
        """AuthHelper.get_token Raises TokenNotFound when Authorization header value not starts with "Bearer" """
        with self.assertRaises(MustBeBearerToken):
            self.auth.get_token({"Authorization": "Invalid Start"})

    def test_04_get_token_error_malformed(self):
        """AuthHelper.get_token Raises TokenNotFound when Authorization header value is malformed"""
        with self.assertRaises(MustStartWithBearer):
            self.auth.get_token({"Authorization": "Invalid parts amount"})

    def test_05_get_token_invalid_headers(self):
        """AuthHelper.get_token Raises TokenNotFound when provided headers isn't an dict instance"""
        with self.assertRaises(TokenNotFound):
            self.auth.get_token([])
        with self.assertRaises(TokenNotFound):
            self.auth.get_token("invalid-headers")
        with self.assertRaises(TokenNotFound):
            self.auth.get_token(1)

    def test_06_get_claims_unable_decode_token(self):
        """AuthHelper.claims Raises UnableDecodeToken when provided token it's invalid"""
        with self.assertRaises(UnableDecodeToken):
            self.auth.claims("invalid-token")
        with self.assertRaises(UnableDecodeToken):
            self.auth.claims(1)

    def test_07_get_claims_unable_decode_token(self):
        """AuthHelper.claims Raises RequiredKeyNotFound when provided not contains claims"""
        with self.assertRaises(ClaimsNotFound):
            self.auth.claims(self.invalid_token_no_claims)

    def test_08_get_claims_returns_dict(self):
        """AuthHelper.claims returns an dict instance"""
        _claims = self.auth.claims(self.valid_token)
        self.assertEqual(type(_claims), dict)
        self.assertEqual(_claims, PAYLOAD)

    def test_09_get_namespaces_no_scopes_found(self):
        """AuthHelper.namespaces when provided token not contain "scope" in claims"""
        with self.assertRaises(ScopesNotFound):
            self.auth.namespaces(self.invalid_token_no_scopes)

    def test_10_get_scopes_on_success(self):
        """AuthHelper.scopes returns an list of str"""
        scopes = self.auth.scopes(self.valid_token)
        self.assertEqual(type(scopes), list)
        self.assertEqual(scopes, ["test:testing"])

    def test_11_get_namespaces_returns_an_list_of_str(self):
        """AuthHelper.namespaces on success return an list of str"""
        namespaces = self.auth.namespaces(self.valid_token)
        self.assertEqual(type(namespaces), list)
        self.assertEqual(namespaces, ["test"])

    def test_12_has_required_scopes_insufficient_permission(self):
        """AuthHelper.has_required_scopes raises InsufficientScopes when not has all required scopes"""
        with self.assertRaises(InsufficientScopes):
            self.auth.has_required_scopes(
                self.valid_token, "has-not:this-scope", require_all=False
            )
        with self.assertRaises(InsufficientScopes):
            self.auth.has_required_scopes(
                self.valid_token, ["test:testing", "has-not:this-scope"]
            )

    def test_13_has_required_scopes_granted(self):
        """AuthHelper.has_required_scopes granted permission"""
        has_all_required_scopes = self.auth.has_required_scopes(
            self.valid_token, "test:testing"
        )
        self.assertEqual(has_all_required_scopes, True)
        has_at_least_one_required_scopes = self.auth.has_required_scopes(
            self.valid_token, ["test:testing", "other:scope"], require_all=False
        )
        self.assertEqual(has_at_least_one_required_scopes, True)

    def test_14_has_required_namespace_insufficient_permission(self):
        """AuthHelper.has_namespace returns false if token not contains required Namespace"""
        has_required_namespace = self.auth.has_namespace(
            "has_not_this_namespace", self.valid_token
        )
        # self.assertEqual(has_required_namespace, False)

    def test_15_has_required_scopes_granted(self):
        """AuthHelper.has_namespace granted permission"""
        has_required_namespace = self.auth.has_namespace("test", self.valid_token)
        self.assertEqual(has_required_namespace, True)



from django.http import HttpResponse
import json


from tokko_auth.decorators import has_permission
from tokko_auth.models import User
from tokko_auth import settings


def auth_status(request):
    user = (
        json.loads(f"{request.user}")
        if isinstance(request.user, User)
        else f"{request.user}"
    )
    response = {
        "user": user,
        "settings": {
            "is_enable": not settings.AUTH_FULL_DISABLED,
            "fail_safe": settings.AUTH_FAIL_SAFE_ENABLE,
            "api_namespace": settings.AUTH_API_NAMESPACE,
            "userinfo": settings.AUTH_USERINFO_ENABLE,
            "auth_domain": settings.AUTH0_DOMAIN,
            "user_successfully_loaded": hasattr(request.user, "claims"),
        },
    }
    return HttpResponse(json.dumps(response, indent=4), content_type="application/json")


@has_permission('test:testing', has_all=True)
def protected_view(request):
    response = {"message": "Authorized!"}
    return HttpResponse(json.dumps(response, indent=4), content_type="application/json")



from urllib.parse import urlencode
from dataclasses import dataclass
from typing import List
import logging
import json

from requests.exceptions import ConnectionError, HTTPError
import requests

from tokko_auth.future.encodings import ErrorDataType

logger = logging.getLogger(__name__)
API_HOME_URL = "https://9fv0mse9dc.execute-api.us-east-1.amazonaws.com/dev-permission-srv"


class Validator:

    __errors__: List[ErrorDataType] = []
    fail_safe: bool = True
    auto_init: bool = True

    def validate(self):

        validators = [m for m in dir(self) if m.startswith("validate_")]
        validators = [getattr(self, v) for v in validators if callable(getattr(self, v))]

        for validator in validators:
            try:
                validator()
            except Exception as e:
                if not self.fail_safe:
                    raise e
                self.add_errors(message=f"{validator.__name__}.{validator.__doc__}.{e}", exception=e)

        return len(self.errors) == 0

    def is_valid(self):
        try:
            if not self.validate():
                raise ValueError(", ".join([f"{err}" for err in self.errors]))

            return True

        except Exception as e:
            logger.error(e)
            if not self.fail_safe:
                raise e

            return False

    @property
    def errors(self) -> list:
        return self.__errors__

    def add_errors(self, message, **extras):

        exception = extras.get("exception")
        status = extras.get("status")
        self.__errors__.append(ErrorDataType(exception=exception,
                                             message=message,
                                             status=status))

    def __post_init__(self):
        if self.auto_init:
            self.validate()


@dataclass(init=True)
class User(Validator):
    user: str = None
    permission: list = None

    def validate_user_id_not_empty(self):
        if not self.user_id:
            raise ValueError("USER_ID is an required argument")

    def __render_permission__(self):
        ...

    def permissions(self) -> list:
        perms = []
        for permission in self.permission:
            if permission and isinstance(permission, str):
                namespace, action = permission.split(":")
                if all([namespace, action]):
                    perms.append(Permission(namespace=namespace, action=action))
        return perms

    def as_dict(self) -> dict:
        return {
            "user_id": self.user,
            "permissions": self.permission
        }

    def __str__(self) -> str:
        return json.dumps(self.as_dict(), indent=4)


@dataclass(init=True)
class Permission(Validator):

    namespace: str = ""
    action: str = ""
    description: str = ""
    permission: str = ""
    auto_test: bool = True

    def validate_namespace_not_empty(self):
        if not self.namespace:
            raise ValueError("NAMESPACE argument is required.")

        return True

    def validate_action_not_empty(self):
        if not self.action:
            raise ValueError("ACTION argument is required.")

        return True

    @property
    def claim(self) -> str:
        return f"{self.namespace}:{self.action}"

    def __post_init__(self):
        if self.auto_test:
            self.validate()

    def __str__(self) -> str:
        return self.claim


class API:

    def __init__(self, **options):
        headers = options.get("headers", {})
        self.token = options.get("token")
        self.session = requests.Session()
        if self.token:
            headers.update({"Authorization": f"Bearer {self.token}"})
        headers.update({"Content-Type": "application/json"})
        self.session.headers.update(headers)

    def run_query(self, url: str, method: str = None, **data):

        method = method or "get"
        allowed_methods = ["post", "get", "put", "delete"]
        if method.lower() not in allowed_methods:
            raise ValueError("Unsupported method")

        try:
            if method.lower() == "get" and data:
                raise IOError("GET not supports DATA")
            _method_fn = getattr(self.session, method.lower())
            r = _method_fn(url, data=json.dumps(data))
            if not r.status_code == 200:
                r.raise_for_status()
            if "application/json" not in r.headers.get("Content-Type", ""):
                raise TypeError(f"API '{self.__name__}' Error. Unsupported response ContentType")

            return r.json()

        except (ConnectionError, HTTPError) as conn_err:
            raise IOError(f"API '{type(self).__name__}' Error. Connection Error {conn_err}")


class PermissionManager(API):

    def add(self, namespace: str, action: str, description: str = None):
        try:
            res = self.run_query(url=f"{API_HOME_URL}/permissions", method='post',
                                 # Data
                                 namespace=namespace,
                                 action=action,
                                 description=description or "")
            return Permission(**res)
        except Exception as e:
            logger.error(f"{e}")

    def delete(self, namespace: str, action: str):

        try:
            return self.run_query(url=f"{API_HOME_URL}/permissions", method="delete",
                                  # Data
                                  namespace=namespace,
                                  action=action)
        except Exception as e:
            logger.error(f"{e}")

    # def set(self, pid: str, namespace: str = None, action: str = None):
    #
    #     try:
    #         return self.run_query(url=f"{API_HOME_URL}/permissions", method="put",
    #                               # Data
    #                               pid=pid, namespace=namespace, action=action)
    #     except Exception as e:
    #         logger.error(f"{e}")

    def get(self, pid: str):

        url = f"{API_HOME_URL}/permissions?{urlencode(dict(pid=pid))}/"
        res = self.run_query(url=url, method='get')
        if res:
            return Permission(**res[0])
        raise IOError("Not Found")

    def list(self, **filters):

        try:
            permissions_url = f"{API_HOME_URL}/permissions"
            if filters:
                permissions_url = f"{permissions_url}?{urlencode(filters)}"
            res = self.run_query(url=permissions_url, method='get')
            if res:
                return [Permission(**permission) for permission in res]
            return []

        except Exception as e:
            logger.error(f"{e}")


class UserPermissionManager(API):

    def grant(self, user_id, namespace, action):

        try:
            permissions_url = f"{API_HOME_URL}/permissions/{user_id}/{namespace}:{action}"
            result = self.run_query(url=permissions_url, method='post')
            return result
        except Exception as e:
            logger.error(f"{e}")

    def revoke(self, user_id, namespace, action):

        try:
            permissions_url = f"{API_HOME_URL}/permissions/{user_id}/{namespace}:{action}"
            result = self.run_query(url=permissions_url, method='delete')
            return result
        except Exception as e:
            logger.error(f"{e}")

    def get(self, user_id):

        try:
            permissions_url = f"{API_HOME_URL}/permissions/{user_id}"
            user_permissions = self.run_query(url=permissions_url, method='get')
            return User(**user_permissions)
        except Exception as e:
            logger.error(f"{e}")


class PermissionProvider:

    permissions_class = PermissionManager
    user_permission_class = UserPermissionManager

    def __init__(self, **options):
        token = options.get("token")
        headers = options.get("headers", {})
        _permission_class = self.permissions_class
        _user_permission_class = self.user_permission_class

        self.permissions = _permission_class(
            headers=headers,
            token=token
        )
        self.users = _user_permission_class(
            headers=headers,
            token=token
        )



"""
# Tutorial I: Dame un token!

En este tutorial vamos a ver como rapidamente podemos hacernos con un JWT utilizando el modulo **TokenFabrik**

```python
from tokko_auth.future.testing import TokenFabrik

token_fabric = TokenFabrik()

print(token_fabric.token)
```

## Results

### Token
```
eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiIsImtpZCI6Im5vLWtpZC1wcm92aWRlZCJ9.
eyJzY29wZSI6ImVtYWlsIHByb2ZpbGUgb3BlbmlkIiwicGVybWlzc2lvbnMiOiIiLCJpc
3MiOiJodHRwczovL3Rlc3RpbmcuY29tL3NlcnZpY2UiLCJhdWQiOiJodHRwczovL3Rlc3
RpbmcucHJvdmlkZXIuY29tLyIsImlhdCI6MTU4NzY4MjYwMSwiZXhwIjoxNTg3NzY5MDA
xLCJndHkiOiJjbGllbnQtY3JlZGVudGlhbHMifQ.gXa5-T3azYOXAWShV2l8unQj6plLD
yrrZZevWwfIsb8eDdYBJLu08jS5fcH7rTtXUMxRd3HNHhQFMkTNJJyZnZ5_uYE54rZZRd
2C3EsgjU_0em6KFIFnPF-WraQuzgWQ8xiOmzzpdUiv6UYQLrBXkGQlQnk-pQuyC8Y_3MK
b0qQ
```

_Nota Importante:_ JTW de ejemplo. El JWT contiene fechas en su interior y
por muy posiblemente, el token generado no sea ESTRICTAMENTE igual al expuesto.

### Headers

```
{
  "typ": "JWT",
  "alg": "RS256",
  "kid": "no-kid-provided"
}
```

### Payload

```
{
  "scope": "email profile openid",
  "permissions": "",
  "iss": "https://testing.com/service",
  "aud": "https://testing.provider.com/",
  "iat": 1587682601,
  "exp": 1587769001,
  "gty": "client-credentials"
}
```

---

# Tutorial II: Crear Tokens Customizados

**TokenFabrik** incluso permite la creacion de JWTs customizados. Es posible agregar nuevos
headers, nuevos claims, especificar permisos etc.


```python
from tokko_auth.future.testing import TokenFabrik

token_fabric = TokenFabrik(
    headers={
        "kid": "custom-key-id",
        "extra_headers":
            {
                "custom_header_01": "custom-value-01",
                "custom_header_02": "custom-value-02",
            }},
    extra_claims={
        "my-custom-claim": "custom-claim-value"
    },
    permission=[
        "namespace:action01",
        "namespace:action02",
        "namespace01:action01"
    ]
)
```

## Results

#### Token
```
eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiIsImN1c3RvbV9oZWFkZXJfMDEiOiJjdXN0b20tdmFsdWUtMDEiLCJ
jdXN0b21faGVhZGVyXzAyIjoiY3VzdG9tLXZhbHVlLTAyIiwia2lkIjoiY3VzdG9tLWtleS1pZCJ9.eyJzY29wZ
SI6ImVtYWlsIHByb2ZpbGUgb3BlbmlkIiwicGVybWlzc2lvbnMiOiJuYW1lc3BhY2U6YWN0aW9uMDEsbmFtZXNw
YWNlOmFjdGlvbjAyLG5hbWVzcGFjZTAxOmFjdGlvbjAxIiwiaXNzIjoiaHR0cHM6Ly90ZXN0aW5nLmNvbS9zZXJ
2aWNlIiwiYXVkIjoiaHR0cHM6Ly90ZXN0aW5nLnByb3ZpZGVyLmNvbS8iLCJpYXQiOjE1ODc2ODE5ODgsImV4cC
I6MTU4Nzc2ODM4OCwiZ3R5IjoiY2xpZW50LWNyZWRlbnRpYWxzIiwibXktY3VzdG9tLWNsYWltIjoiY3VzdG9tL
WNsYWltLXZhbHVlIn0.nNLVS8VClhpHlV6ZjN31Z1TcZKuq9uPh5DY5J1ZvmwxMQBL2sLd2jqiGk0l0_7eesWiI
rd9EeLCYhEnHq6m_bRgI8Uyy7YXs5X3_0GfzVkSNPFxOhoxuj5ZP11uk77UZBWGa7Dcikux19tccsk-fnEJbQGA
dRPx-sbHpmrV1Wx8
```

_Nota Importante:_ JWT de ejemplo. El JWT contiene fechas en su interior y por muy
posiblemente, el token generado no sea ESTRICTAMENTE igual al expuesto.

#### Headers
```
{
  "typ": "JWT",
  "alg": "RS256",
  "custom_header_01": "custom-value-01",
  "custom_header_02": "custom-value-02",
  "kid": "custom-key-id"
}
```

#### Payload:
```
{
  "scope": "email profile openid",
  "permissions": "namespace:action01,namespace:action02,namespace01:action01",
  "iss": "https://testing.com/service",
  "aud": "https://testing.provider.com/",
  "iat": 1587681988,
  "exp": 1587768388,
  "gty": "client-credentials",
  "my-custom-claim": "custom-claim-value"
}
```
"""

from dataclasses import dataclass
import json
import os

from Crypto.PublicKey import RSA
from arrow import get as _, now
from typing import Union, List
from jose import jwt
import jwcrypto.jwk as jwk

ListOrDict = Union[dict, list]


@dataclass
class TokenHeaders:
    kid: Union[int, str] = "no-kid-provided"
    extra_headers: dict = None

    def as_dict(self) -> dict:
        headers = self.extra_headers or {}

        headers.update({"kid": self.kid})

        return headers

    def __str__(self) -> str:
        return json.dumps(self.as_dict(), indent=4)


@dataclass
class Payload:
    """Payload interface"""

    # Testing constants
    DEFAULT_SCOPE = "email profile openid"
    PERMISSION = []
    DEFAULT_AUD = "https://testing.provider.com/"
    DEFAULT_GTY = "client-credentials"
    DEFAULT_ISS = "https://testing.com/service"

    __epoch__ = _("1970-01-01")
    __now__ = now()

    scope: str = DEFAULT_SCOPE
    permissions: List[str] = None
    expiration_minutes: int = 24 * 60  # Token expires in 1 day.
    aud: str = DEFAULT_AUD
    iss: str = DEFAULT_ISS
    gty: str = DEFAULT_GTY

    extra_claim: dict = None

    @property
    def iat(self) -> int:
        return round((self.__now__ - self.__epoch__).total_seconds())

    @property
    def exp(self) -> int:
        _n = self.__now__
        return round((_n.shift(minutes=self.expiration_minutes) - self.__epoch__).total_seconds())

    def as_dict(self) -> dict:
        permissions = ",".join(self.permissions or [])
        claims = {
            "scope": self.scope,
            "permissions": permissions,
            "iss": self.iss,
            "aud": self.aud,
            "iat": self.iat,
            "exp": self.exp,
            "gty": self.gty,
        }
        claims.update(self.extra_claim or {})

        return claims

    def __str__(self) -> str:
        return json.dumps(self.as_dict(), indent=4)


FilePath = str
PrivateKey = str
RSAKey = Union[FilePath, PrivateKey]


@dataclass(init=True)
class RSAIDManager:

    private_key_pem: RSAKey = None
    load_from_file: bool = False
    root_folder: FilePath = "./"
    pk_bits: int = 1024

    __public_key__ = None
    __private_key__ = None

    @property
    def public_key(self):
        return self.__public_key__

    @property
    def private_key(self):
        return self.__private_key__

    @property
    def jwks(self):
        _keys = json.loads(jwk.JWK.from_pem(self.public_key).export())
        _keys.update({"use": "access"})
        return {
            "keys": [_keys]
        }

    @staticmethod
    def load_pem_from_file(pem_filename, root_folder="./"):

        pem_file = os.path.join(pem_filename, root_folder)
        with open(pem_file, "r") as pkf:

            private_key = pkf.read()

        return private_key

    def load_rsa(self):

        if not self.private_key_pem:
            rsa_key = RSA.generate(bits=self.pk_bits)

        else:
            if not isinstance(self.private_key, str):
                raise TypeError
            rsa_key = RSA.importKey(self.private_key)

        self.__private_key__ = rsa_key.export_key().decode()
        self.__public_key__ = rsa_key.publickey().exportKey('PEM')

        return self.private_key

    def __post_init__(self):
        if self.load_from_file:
            self.private_key_pem = self.load_pem_from_file(
                pem_filename=self.private_key_pem,
                root_folder=self.root_folder,
            )
        self.load_rsa()


@dataclass
class TokenFabrik:

    pk_bits: int = 1024
    private_key_pem: str = None
    load_pem_from_file: bool = False

    headers: dict = None
    payload: dict = None
    algorithm: str = "RS256"
    permission: list = None
    extra_claims: dict = None
    auth_header_key_name = "Authorization"

    __token__ = None
    __rsa_id__: RSAIDManager = None
    __initialized__ = False

    @staticmethod
    def build_token(payload: ListOrDict, rsa_id: str, algorithm: str, headers: dict):

        return jwt.encode(
            payload,
            key=rsa_id,
            algorithm=algorithm,
            headers=headers,
        )

    def __initialize_headers__(self):

        headers = self.headers or {}
        headers.update({"use": "access"})

        self.headers = TokenHeaders(kid=self.jwks["keys"][0]["kid"],
                                    extra_headers=headers).as_dict()

    def __initialize_payload__(self):

        payload = self.payload or {}
        self.payload = Payload(permissions=self.permission,
                               extra_claim=self.extra_claims,
                               aud=os.environ.get("API_AUDIENCE"),
                               **payload).as_dict()

    def __initialize_rsa_id__(self):

        self.__rsa_id__ = RSAIDManager(private_key_pem=self.private_key_pem,
                                       load_from_file=self.load_pem_from_file,
                                       pk_bits=self.pk_bits)

    def initialize(self):

        if not self.__initialized__:
            self.__initialize_rsa_id__()
            self.__initialize_headers__()
            self.__initialize_payload__()
            self.__initialized__ = True

    def __post_init__(self):

        self.initialize()

    def authorization_header(self) -> dict:

        return {self.auth_header_key_name: self.bearer}

    @property
    def private_key(self):

        return self.__rsa_id__.private_key

    @property
    def public_key(self):

        return self.__rsa_id__.public_key

    @property
    def jwks(self):

        return self.__rsa_id__.jwks

    @property
    def bearer(self):

        return f"Bearer {self.token}"

    @property
    def token(self):

        if not self.__token__:
            self.__token__ = self.build_token(payload=self.payload,
                                              headers=self.headers,
                                              rsa_id=self.private_key,
                                              algorithm=self.algorithm)
        return self.__token__




from dataclasses import dataclass


@dataclass(init=True)
class ErrorDataType:

    message: str
    exception: Exception = None
    status: int = None

    def __str__(self):

        exception = self.exception or Exception(self.message)
        status = self.status or "-"

        return f"{self.message} - " \
               f"{type(exception).__name__} " \
               f"[{status}]"




from unittest import TestCase

from tokko_auth.future.encodings import ErrorDataType


class EncodingsTestCase(TestCase):

    def test_01_error_data_type_print(self):

        edt_a = ErrorDataType(**{"message": "testing", "exception": TypeError(), "status": 500})
        edt_b = ErrorDataType(message="testing")

        self.assertEquals(f"{edt_a}", "testing - TypeError [500]")
        self.assertEquals(f"{edt_b}", "testing - Exception [-]")


class ToolsPermissionTestCase(TestCase):
    ...


class ToolsTestingTestCase(TestCase):
    ...



from django.conf.urls import url

from tokko_auth.views import auth_status, protected_view
from tokko_auth import settings

# Other auth path
urlpatterns = []

if settings.SAMPLES_ARE_ENABLE:
    urlpatterns += [
        # Samples
        url("status/", auth_status),
        url("protected/", protected_view),
    ]
